#include <stdlib.h>
#include <stdio.h>

#include "deque_of.h"

void deque_init(deque_t* dq, int q_size, int data_size) {
  int i;

  dq->m_QUEUE_SIZE = q_size + 2;
  dq->m_elems = (int*) malloc((dq->m_QUEUE_SIZE) * sizeof(int));
  for (i = 0; i < dq->m_QUEUE_SIZE / 2; i++) {
    set_ctr(&dq->m_elems[i], 0);
    set_val(&dq->m_elems[i], LN);
  }
  for (; i < dq->m_QUEUE_SIZE; i++) {
    set_ctr(&dq->m_elems[i], 0);
    set_val(&dq->m_elems[i], RN);
  }

  dq->m_DATA_SIZE = data_size;
  dq->m_data = (int*) malloc(dq->m_DATA_SIZE * sizeof(int));
  for (i = 0; i < dq->m_DATA_SIZE; i++) {
    dq->m_data[i] = rand();
  }
}

void deque_destroy(deque_t* dq) {
  free(dq->m_elems);
  free(dq->m_data);
}

int get_ctr(int elem) {
  return elem >> 16;
}

int get_val(int elem) {
  return elem & 0x0000FFFF;
}

void set_ctr(int* elem, unsigned short val) {
  int sval = val << 16;
  *elem = *elem & 0x0000FFFF;
  *elem = *elem | sval;
}

void set_val(int* elem, unsigned short val) {
  *elem = *elem & 0xFFFF0000;
  *elem = *elem | val;
}

int enqueueLeft(deque_t* dq, unsigned short elem) {
  int k, prev, cur, _old, _new;

  dq_doWork(dq);
  while (1) {
    k = oracle(dq, 0);
    prev = dq->m_elems[k + 1];
    cur = dq->m_elems[k];
    if (get_val(prev) != LN && get_val(cur) == LN) {
      if (k == 0) {
        return 0;
      }
      _old = prev;
      _new = prev;
      set_ctr(&_new, get_ctr(_new) + 1);
      CAS(dq->m_elems[k + 1], _old, _new);
      if (_old == _new) {
        _old = cur;
        _new = cur;
        set_ctr(&_new, get_ctr(_new) + 1);
        set_val(&_new, elem);
        CAS(dq->m_elems[k], _old, _new);
        if (_old == _new) {
          return 1;
        } 
      }
    }
  }
}

int enqueueRight(deque_t* dq, unsigned short elem) {
  int k, prev, cur, _old, _new;

  dq_doWork(dq);
  while (1) {
    k = oracle(dq, 1);
    prev = dq->m_elems[k - 1];
    cur = dq->m_elems[k];
    if (get_val(prev) != RN && get_val(cur) == RN) {
      if (k == dq->m_QUEUE_SIZE - 1) {
        return 0;
      }
      _old = prev;
      _new = prev;
      set_ctr(&_new, get_ctr(_new) + 1);
      CAS(dq->m_elems[k - 1], _old, _new);
      if (_old == _new) {
        _old = cur;
        _new = cur;
        set_ctr(&_new, get_ctr(_new) + 1);
        set_val(&_new, elem);
        CAS(dq->m_elems[k], _old, _new);
        if (_old == _new) {
          return 1;
        } 
      }
    }
  }
}

int dequeueLeft(deque_t* dq, unsigned short* elem) {
  int k;
  unsigned int cur, next, _old, _new;

  dq_doWork(dq);
  while (1) {
    k = oracle(dq, 0);
    cur = dq->m_elems[k + 1];
    next = dq->m_elems[k];
    if (get_val(cur) != LN && get_val(next) == LN) {
      if (get_val(cur) == RN && dq->m_elems[k + 1] == cur) {
        *elem = NULL;
        return 0;
      }
      _old = next;
      _new = next;
      set_ctr(&_new, get_ctr(_new) + 1);
      CAS(dq->m_elems[k], _old, _new);
      if (_old == _new) {
        _old = cur;
        _new = cur;
        set_ctr(&_new, get_ctr(_new) + 1);
        set_val(&_new, LN);
        CAS(dq->m_elems[k + 1], _old, _new);
        if (_old == _new) {
          *elem = (unsigned short) get_val(cur);
          return 1;
        }
      }
    }
  }
}

int dequeueRight(deque_t* dq, unsigned short* elem) {
  int k;
  unsigned int cur, next, _old, _new;

  dq_doWork(dq);
  while (1) {
    k = oracle(dq, 1);
    cur = dq->m_elems[k - 1];
    next = dq->m_elems[k];
    if (get_val(cur) != RN && get_val(next) == RN) {
      if (get_val(cur) == LN && dq->m_elems[k - 1] == cur) {
        *elem = NULL;
        return 0;
      }
      _old = next;
      _new = next;
      set_ctr(&_new, get_ctr(_new) + 1);
      CAS(dq->m_elems[k], _old, _new);
      if (_old == _new) {
        _old = cur;
        _new = cur;
        set_ctr(&_new, get_ctr(_new) + 1);
        set_val(&_new, RN);
        CAS(dq->m_elems[k - 1], _old, _new);
        if (_old == _new) {
          *elem = (unsigned short) get_val(cur);
          return 1;
        }
      }
    }
  }
}

void dq_doWork(deque_t* dq) {
  int i, total = 0;
  for (i = 0; i < dq->m_DATA_SIZE; i++) {
    total += dq->m_data[i];
  }
}

int oracle(deque_t* dq, int right) {
  int i;
  if (right) {
    for (i = 1;
         get_val(dq->m_elems[i]) != RN && i < dq->m_QUEUE_SIZE - 1;
         i++) {
    }
  } else {
    for (i = dq->m_QUEUE_SIZE - 2;
         get_val(dq->m_elems[i]) != LN && i >= 1; i--) {}
  }
  return i;
}
