#ifndef _DEQUE_H_
#define _DEQUE_H_

#define RN 0x0000FFFF
#define LN 0x0000FFFE

/* CAS macro
   semantics:  
    If (a == b) then a = c
    c = OLD value of a
*/    
#define CAS(a,b,c)   \
asm volatile(                           \
  "mov %1,%%l2\n"        \
  "mov %2,%%l1\n"        \
  "mov %0,%%l3\n"        \
  "cas [%%l2],%%l1,%%l3\n" \
  "mov %%l3,%3\n"    \
  :"=r"(c) \
  : "r"(&a), "r"(b),"r"(c)       \
  :"l1","l2","l3"                        \
);

typedef struct {
  int m_QUEUE_SIZE;
  int m_DATA_SIZE;

  unsigned int* m_elems;
  int* m_data;
} deque_t;

// must do srand elsewhere if desired for constructors
void deque_init(deque_t* dq, int qSize, int dataSize);
void deque_destroy(deque_t* dq);

// idx in the top bits
// value in the bottom ones
int get_ctr(int elem);
int get_val(int elem);
void set_ctr(int* elem, unsigned short val);
void set_val(int* elem, unsigned short val);

// return true if successful: there was room
int enqueueLeft(deque_t* dq, unsigned short elem);
int enqueueRight(deque_t* dq, unsigned short elem);
int dequeueLeft(deque_t* dq, unsigned short* elem);
int dequeueRight(deque_t* dq, unsigned short* elem);
void dq_doWork(deque_t* dq);
int oracle(deque_t* dq, int right);

#endif
