#include <assert.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/processor.h>
#include <sys/procset.h>
#include <sys/types.h>

#include "barrier.h"
#include "deque.h"
#include <magic_services.h>
#include "transaction.h"

#define NUM_OPS 32768
#define NUM_THREADS 8
#define MAX_PROCS 8
#define DEFAULT_QUEUE_SIZE 1024
#define DEFAULT_DATA_SIZE 512

int MAX_THREAD_WORK;
int* data;
int processor_ids[MAX_PROCS] = {0,1,4,5,8,9,10,11}; //{0, 1, 4, 5, 8, 9, 12, 13};
barrier_t bar; // barrier
deque_t* dqp;
pthread_mutex_t dq_lock; // mutex for whole dq

void thread_doWork() {
  int i;

  if (MAX_THREAD_WORK == 0) {
    return;
  }
  int amt = rand() * MAX_THREAD_WORK / RAND_MAX;
  int total = 0;
  for (i = 0; i < amt; i++) {
    total += data[i];
  }
}

void* threadMain(void* my_id) {
  int i;
  int local_sense = 0; // for sense-reversing barrier
  int id = (int) my_id;
  int data;
  // bind thread to a processor, ensuring at most one thread per proc
  //cout << id << " binding..." << endl;
  assert(!processor_bind(P_LWPID, P_MYID, processor_ids[id], NULL));

  if (id == 0) {
    RUBY_MAGIC_CALL(Do_Breakpoint, 0);
  }
  barrier(&bar, &local_sense);
  
  int NUM_ACTIONS = 4;
  int action;
  for (i = 0; i < NUM_OPS / NUM_THREADS; i++) {
    action = rand() * NUM_ACTIONS / RAND_MAX;
    switch (action) {
    case 0:
      BEGIN_TRANSACTION(0);
      enqueueLeft(dqp, 0);
      COMMIT_TRANSACTION(0);
      break;
    case 1:
      BEGIN_TRANSACTION(0);
      enqueueRight(dqp, 1);
      COMMIT_TRANSACTION(0);
      break;
    case 2:
      BEGIN_TRANSACTION(0);
      dequeueLeft(dqp, &data);
      COMMIT_TRANSACTION(0);
      break;
    case 3:
      BEGIN_TRANSACTION(0);
      dequeueRight(dqp, &data);
      COMMIT_TRANSACTION(0);
      break;
    default:
      ;
      //cout << "error!";
    }
    thread_doWork();
  }

  barrier(&bar, &local_sense);
  if (id == 0) {
    RUBY_MAGIC_CALL(Do_Breakpoint, 0);
  }
  return NULL;
}

int main(int argc, char** argv) {
  int i;

  // parse arguments
  if (argc < 3) {
    printf("* USAGE: \n");
    printf("* %s <dequeWork> <threadWork> \n", argv[0]);
    exit(-1);
  }
  srand(time(0));
  dqp = (deque_t*) malloc(sizeof(deque_t));
  deque_init(dqp, DEFAULT_QUEUE_SIZE, atoi(argv[1]));
  MAX_THREAD_WORK = atoi(argv[2]);
  data = (int*) malloc(MAX_THREAD_WORK * sizeof(int));
  for (i = 0; i < MAX_THREAD_WORK; i++) {
    data[i] = rand();
  }

  // set up threads
  pthread_t* threads = (pthread_t*) malloc(NUM_THREADS * sizeof(pthread_t));
  pthread_attr_t attr;
  assert(!pthread_attr_init(&attr));
  assert(!pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM));
  barrier_init(&bar, NUM_THREADS);
  pthread_mutex_init(&dq_lock, NULL);

  // start the threads in the thread_main routine
  for (i = 0; i < NUM_THREADS; i++) {
    assert(!pthread_create(&threads[i], &attr, threadMain, (void*) i));
  }

  // as each thread completes, merge with it
  for (i = 0; i < NUM_THREADS; i++) {
    assert(!pthread_join(threads[i], NULL));
  }

  // clean up
  deque_destroy(dqp);
  free(dqp);
  free(data);
  free(threads);
  pthread_attr_destroy(&attr);
  barrier_destroy(&bar);
  pthread_mutex_destroy(&dq_lock);

  return 0;
}
