#include <assert.h>
#include <iostream.h>
#include <pthread.h>
#include <sys/processor.h>
#include <sys/procset.h>
#include <sys/types.h>

#include "barrier.h"
#include "Deque.h"

int const NUM_OPS = 8;//32768;
int const NUM_THREADS = 8;
int MAX_THREAD_WORK;
int* data;
int const MAX_PROCS = 8;
int processor_ids[MAX_PROCS] = {0, 1, 4, 5, 8, 9, 12, 13}; // cabernet
barrier_t bar; // barrier
Deque<int>* dqp;

void doWork() {
  if (MAX_THREAD_WORK == 0) {
    return;
  }
  int amt = (rand() * (MAX_THREAD_WORK + 1) / RAND_MAX) + 1;
  assert(amt > 0 && amt <= MAX_THREAD_WORK);
  int total = 0;
  for (int i = 0; i < amt; i++) {
    total += data[i];
  }
}

void printData(int const * data, int size) {
  int i;
  cout << "[";
  for (i = 0; i < size - 1; i++) {
    cout << data[i] << ", ";
  }
  cout << data[i] << "]" << endl;
}

void test() {
  Deque<int> d(8, 16);
  cout << "left=" << d.getLeft() << ", right=" << d.getRight();
  cout << ", size=" << d.getSize() << endl;

  int const * data = d.getData();
  printData(data, d.getDataSize());

  int i = 0;
  while (d.enqueueLeft(i)) {
    i++;
  };
  cout << "left=" << d.getLeft() << endl;

  i = 0;
  while (d.enqueueRight(i)) {
    i++;
  }
  cout << "right=" << d.getRight() << endl;

  int* ip;
  while (d.dequeueLeft(ip)) {
    cout << "*i=" << *ip << endl;
  }
  cout << "left=" << d.getLeft() << ", right=" << d.getRight() << endl;

  i = 0;
  while (d.enqueueLeft(i)) {
    i++;
  };
  cout << "left=" << d.getLeft() << endl;

  while (d.dequeueRight(ip)) {
    cout << "*i=" << *ip << endl;
  }
  cout << "left=" << d.getLeft() << ", right=" << d.getRight() << endl;
}

void* threadMain(void* my_id) {
  int local_sense = 0; // for sense-reversing barrier
  int id = (int) my_id;
  int data;
  // bind thread to a processor, ensuring at most one thread per proc
  //cout << id << " binding..." << endl;
  assert(!processor_bind(P_LWPID, P_MYID, processor_ids[id], NULL));
  barrier(&bar, local_sense);
  
  int const NUM_ACTIONS = 4;
  int action;
  for (int i = 0; i < NUM_OPS / NUM_THREADS; i++) {
    action = rand() * NUM_ACTIONS / RAND_MAX;
    switch (action) {
    case 0:
      dqp->enqueueLeft(0);
      break;
    case 1:
      dqp->enqueueRight(1);
      break;
    case 2:
      dqp->dequeueLeft(&data);
      break;
    case 3:
      dqp->dequeueRight(&data);
      break;
    default:
      cout << "error!";
    }
    doWork();
  }
}

int main(int argc, char** argv) {
  // parse arguments
  if (argc < 3) {
    cout << "* USAGE: " << endl;
    cout << "* " << argv[0] << " <dequeWork> <threadWork> " << endl;
    exit(-1);
  }
  srand(time(0));
  dqp = new Deque<int>(Deque<int>::DEFAULT_QUEUE_SIZE, atoi(argv[1]));
  MAX_THREAD_WORK = atoi(argv[2]);
  data = new int[MAX_THREAD_WORK];
  for (int i = 0; i < MAX_THREAD_WORK; i++) {
    data[i] = rand();
  }

  // set up threads
  pthread_t* threads = new pthread_t[NUM_THREADS];
  pthread_attr_t attr;
  assert(!pthread_attr_init(&attr));
  assert(!pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM));
  barrier_init(&bar, NUM_THREADS);

  // start the threads in the thread_main routine
  for (int i = 0; i < NUM_THREADS; i++) {
    assert(!pthread_create(&threads[i], &attr, threadMain, (void*) i));
  }

  // as each thread completes, merge with it
  for (int i = 0; i < NUM_THREADS; i++) {
    assert(!pthread_join(threads[i], NULL));
  }

  // clean up
  delete dqp;
  delete[] data;
  delete[] threads;
  pthread_attr_destroy(&attr);
  barrier_destroy(&bar);

  return 0;
}
