/* test-cas.c
  
    Test program to demonstrate a compile and swap atomic primitive.
    provided by the sparc architecture. 
    Each thread does an atomic fetch-and-increment on a global variable
    thus avoiding the use of a lock. 

    compile:
      gcc -o test_cas -lpthread -Wa,-xarch=v8plusa test-cas.c
      
    Run:
      ./test_cas <num_threads>
*/                

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

/* CAS macro
   semantics:  
    If (a == b) then a = c
    c = OLD value of a
*/    
#define CAS(a,b,c)\
asm volatile(\
 "mov %0,%%l2\n" \
 "mov %1,%%l1\n"  \
 "mov %2,%%l3\n" \
 "cas [%%l2],%%l1,%%l3\n" \
             "mov %%l3, %3\n" \
             :"=r"(c) \
 : "r"(&a), "r"(b),"r"(c)\
 :"l1","l2","l3"\
);

int count = 0;
    
void slave(void *t){
    int myid;
    int inc_count;
        
    while (1){
        /* Get the value of the global variable */            
        myid = count;
        inc_count = myid + 1;
         
        /* Try to write the incremented value into the global variable. This
             will succeed only if the global variable was not overwritten in
             the meanwhile by another thread */
        CAS(count, myid, inc_count)
            if (myid == inc_count)    
                /* CAS succeeded */
                break;
    }
        
    printf("Thread Id: %d\n",myid);
}                    
            
         
        
int main(int argc, char **argv){
    pthread_t *threads;
    pthread_attr_t pthread_custom_attr;
    int num_threads;
    int i;
        
    if (argc != 2){
        printf("Usage: %s <num_threads>\n",argv[0]);
        exit(1);
    }
        
    num_threads = atoi(argv[1]);
    threads = (pthread_t *)calloc(num_threads,sizeof(pthread_t));
    pthread_attr_init(&pthread_custom_attr);

    for ( i = 0; i < num_threads; i++){
        pthread_create(&threads[i], &pthread_custom_attr, slave, NULL);
    }

    for ( i = 0; i < num_threads; i++){
        pthread_join(threads[i],NULL);
    }
}        
                                
